#uses "fwUnitTestComponentAsserts.ctl"
#uses "test/fwRPC/utils.ctl"

#uses "fwRPC/fwRPC_msgBroker.ctl"
#uses "fwRPC/fwRPC_msgClient.ctl"
#uses "fwRPC/fwRPC_common.ctl"
#uses "CtrlSerialization"
#uses "fwRPC/fwRPC_RingBuffer.ctl"

#uses "CtrlZlib"


private global const string _delegate = "_mockdelegate";
private global const string _cdelegate = "_mockClientDelegate";
private global dyn_dyn_mixed _msgs;
private global dyn_dyn_mixed _cmsgs;
private global string _rbDp;
private global string _clientname = "x";
private global anytype _ctx;


void test_fwRPC_msgBroker_setupSuite() {
  _BEGIN_SUITE("msg broker");

  _rbDp = fwRPC_RBMBX_DPT + "/" + fwRPC_DIST_TAG_LOCAL + "_" +
          fwRPC_DIST_TAG_CLIENT + "_" + _clientname;
}


void test_fwRPC_msgBroker_teardownSuite() {
  _END_SUITE();
}


void test_fwRPC_msgBroker_setup() {
  _BEGIN_TEST();
  dynClear(_msgs);
  dynClear(_cmsgs);

  mapping d; d[fwRPC_DISTMSG_TYPE_ACKNOWLEDGE] = _cdelegate;
  int err = fwRPC_msgClient_initialize(_ctx, true, fwRPC_DIST_TAG_CLIENT, _clientname, d);
  assertEqual(0, err);
  delay(0,500);
}


void test_fwRPC_msgBroker_teardown() {
  int err = fwRPC_msgClient_stop(_ctx);
  if (fwRPC_msgBroker_isRunning()) fwRPC_msgBroker_stop();
  delay(1);
}


//------------------------------------------------------------------------------


void _mockdelegate(string project, string name, anytype msg) {
  dynAppend(_msgs, makeDynMixed(project, name, msg));
}


void _mockClientDelegate(anytype ctx, string project, string name, anytype msg) {
  dynAppend(_cmsgs, makeDynMixed(project, name, msg));
}


void _mockNewMbxCB(string project, string name) {
  dynAppend(_msgs, makeDynMixed(project, name));
}


void _getLastClientMessage(
  string &id,
  string &type,
  string &proj,
  string &name,
  anytype &payload
) {
  string val;
  int err = fwRPC_RingBuffer_read(_rbDp + ".in", val);
  assertEqual(0, err, "Error reading DP: " + _rbDp + ".in");

  string buffer;
  fwRPC_common_unpackMessage(val, id, type, proj, name, buffer);
  deserializeFromXml(buffer, payload);
}


//------------------------------------------------------------------------------


void test_fwRPC_msgBroker_listens_valid_param() {
  assertEqual(0, fwRPC_msgBroker_listen(true, ""));
  assertEqual(0, fwRPC_msgBroker_stop());
}


void test_fwRPC_msgBroker_stop_disconnects() {
  int err;
  
  fwRPC_msgBroker_initialize();

  fwRPC_msgBroker_accept(
    "custom",
    _delegate,
    fwRPC_DIST_TAG_CLIENT
  );

  err = fwRPC_msgBroker_listen(true, "");
  assertEqual(0, err);

  err = fwRPC_msgBroker_stop();
  assertEqual(0, err, "should successfully disconnect");
  delay(0, 500);

  fwRPC_msgClient_sendRequest(_ctx, "custom", "neverrecv");

  delay(0, 500);

  assertEqual(0, dynlen(_msgs), "should not receive msg");
}


void test_fwRPC_msgBroker_delegates_accepted_path() {
  int err;

  fwRPC_msgBroker_initialize();

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_ACKNOWLEDGE,
    _delegate,
    fwRPC_DIST_TAG_CLIENT
  );

  err = fwRPC_msgBroker_listen(true, "");
  assertEqual(0, err, "error listen");

  fwRPC_msgClient_sendRequest(_ctx, fwRPC_DISTMSG_TYPE_ACKNOWLEDGE, "a");

  delay(0,800);

  assertEqual(1, dynlen(_msgs), "should have msg");
  dyn_mixed msg = _msgs[1];
  assertEqual(fwRPC_common_cleanProjectName(getSystemName()), msg[1]);
  assertEqual("loc_clnt_x", msg[2]);
  assertEqual("a", msg[3]);
}


void test_fwRPC_msgBroker_delegates_does_not_delegate_unaccepted() {
  int err;

  fwRPC_msgBroker_initialize();

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_ACKNOWLEDGE,
    _delegate,
    fwRPC_DIST_TAG_WORKER
  );

  err = fwRPC_msgBroker_listen(true, "");
  assertEqual(0, err);

  fwRPC_msgClient_sendRequest(_ctx, fwRPC_DISTMSG_TYPE_ACKNOWLEDGE, "a");

  delay(0, 500);

  assertEqual(0, dynlen(_msgs));
}


void test_fwRPC_msgBroker_writes_messages() {
  fwRPC_msgBroker_listen(true);

  int err = fwRPC_msgBroker_writeMessage(
    fwRPC_common_cleanProjectName(getSystemName()), "loc_clnt_x",
    fwRPC_DISTMSG_TYPE_ACKNOWLEDGE, "y"
  );

  assertEqual(0, err);

  delay(0, 500);

  assertEqual(1, dynlen(_cmsgs));
  dyn_mixed msg = _cmsgs[1];
  assertEqual(fwRPC_DIST_BRKR, msg[1], "proj should be brkr");
  assertEqual(fwRPC_DIST_BRKR, msg[2], "name should be brkr");
  assertEqual("y", msg[3]);
}


void test_fwRPC_msgBroker_internal_buffer_if_rb_is_full() {
  const int buffSize = 50;

  //let's stop the client from reading..
  fwRPC_msgClient_stop(_ctx);

  string pname = fwRPC_common_cleanProjectName(getSystemName());

  fwRPC_msgBroker_initialize();
  fwRPC_msgBroker_listen(true);

  int err;
  for (int i=0; i<100; ++i) {
    err = fwRPC_msgBroker_writeMessage(
      pname, "loc_clnt_x",
      fwRPC_DISTMSG_TYPE_ACKNOWLEDGE, i
    );
    assertEqual(0, err);
  }

  //buffer should only contain the first buffSize numbers
  dyn_blob buffContents;
  dpGet(_rbDp + ".in" + fwRPC_RingBuffer_DPE_VALUES, buffContents);
  assertEqual(buffSize, dynlen(buffContents));

  string id, type, proj, name, buffer; anytype payload;
  for (int i=1; i<=buffSize; ++i) {
    string uncompressed;
    gunzip(buffContents[i], uncompressed);
    fwRPC_common_unpackMessage(uncompressed, id, type, proj, name, buffer);
    deserializeFromXml(buffer, payload);
    assertTrue(payload >= 0 && payload < buffSize);
  }

  //trigger drain
  _getLastClientMessage(id, type, proj, name, payload);
  assertEqual(fwRPC_DISTMSG_TYPE_ACKNOWLEDGE, type);
  assertEqual(0, payload);

  delay(0, 500);

  dpGet(_rbDp + ".in" + fwRPC_RingBuffer_DPE_VALUES, buffContents);
  assertEqual(buffSize, dynlen(buffContents));

  //rb should now contain #buffSize
  bool contains = false;
  for (int i=1; i<=buffSize; ++i) {
    string uncompressed;
    gunzip(buffContents[i], uncompressed);
    fwRPC_common_unpackMessage(uncompressed, id, type, proj, name, buffer);
    deserializeFromXml(buffer, payload);
    if (payload == buffSize) {
      contains = true;
      break;
    }
  }

  assertEqual(true, contains, "buffer did not contain new value");

  //let's empty the RB, nums should contain #1 to #50. Drain will be triggered
  for (int i=1; i<=buffSize; ++i) {
    _getLastClientMessage(id, type, proj, name, payload);
    assertEqual(true, payload > 0 && payload <= buffSize);
  }

  //RB should contain #buffsize to #99
  for (int i=1; i<=buffSize-1; ++i) {
    _getLastClientMessage(id, type, proj, name, payload);
    assertEqual(true, payload > buffSize && payload < 100);
  }

  err = fwRPC_RingBuffer_read(_rbDp + ".in", buffer);
  assertEqual(fwRPC_RingBuffer_EEMPTY, err);
}


void test_fwRPC_msgBroker_stores_handle_info() {
  fwRPC_msgBroker_initialize();

  fwRPC_msgBroker_listen(true);

  fwRPC_msgBroker_registerHandle("hdl", "sp", "sn", "tp", "tn");

  string p,n;
  int err = fwRPC_msgBroker_getHandleCaller("hdl", p, n);
  assertEqual(0, err);
  assertEqual("sp", p);
  assertEqual("sn", n);

  err = fwRPC_msgBroker_getHandleCallee("hdl", p, n);
  assertEqual(0, err);
  assertEqual("tp", p);
  assertEqual("tn", n);

  err = fwRPC_msgBroker_getHandleCallee(_clientname, p, n);
  assertEqual(-1, err);
}


void test_fwRPC_msgBroker_stores_handle_target_later() {
  fwRPC_msgBroker_initialize();

  fwRPC_msgBroker_listen(true);

  fwRPC_msgBroker_registerHandle("hdl", "sp", "sn");

  string p, n;
  int err = fwRPC_msgBroker_getHandleCaller("hdl", p, n);
  assertEqual(0, err);
  assertEqual("sp", p);
  assertEqual("sn", n);

  err = fwRPC_msgBroker_getHandleCallee("hdl", p, n);
  assertEqual(-1, err);

  err = fwRPC_msgBroker_setHandleCallee("hdl", "tp", "tn");
  assertEqual(0, err);
  
  err = fwRPC_msgBroker_getHandleCallee("hdl", p, n);
  assertEqual(0, err);
  assertEqual("tp", p);
  assertEqual("tn", n);

  err = fwRPC_msgBroker_setHandleCallee(_clientname, "tp", "tn");
  assertEqual(-1, err);
}


void test_fwRPC_msgBroker_unregsiters_handles() {
  fwRPC_msgBroker_initialize();

  fwRPC_msgBroker_listen(true);

  int err = fwRPC_msgBroker_unregisterHandle("hdl");
  assertEqual(-1, err);

  fwRPC_msgBroker_registerHandle("hdl", "p", "n");

  err = fwRPC_msgBroker_unregisterHandle("hdl");
  assertEqual(0, err);

  string p, n;
  err = fwRPC_msgBroker_getHandleCallee("hdl", p, n);
  assertEqual(-1, err);

  err = fwRPC_msgBroker_getHandleCaller("hdl", p, n);
  assertEqual(-1, err);
}


void test_fwRPC_msgBroker_retrieves_handles_called_by() {
  fwRPC_msgBroker_initialize();

  fwRPC_msgBroker_listen(true);

  dyn_string hdls;

  fwRPC_msgBroker_getHandlesCalledBy(_clientname, "y", hdls);
  assertEqual(0, dynlen(hdls));

  fwRPC_msgBroker_registerHandle("hdl", "p", "n");
  fwRPC_msgBroker_registerHandle("hdl2", "p", "n");
  fwRPC_msgBroker_registerHandle("hdl3", "p",_clientname);
  fwRPC_msgBroker_registerHandle("hdl4",_clientname, "p");

  fwRPC_msgBroker_getHandlesCalledBy("p", "n", hdls);
  assertEqual(2, dynlen(hdls));
  assertTrue(dynContains(hdls, "hdl"));
  assertTrue(dynContains(hdls, "hdl2"));

  fwRPC_msgBroker_getHandlesCalledBy("p",_clientname, hdls);
  assertEqual(1, dynlen(hdls));
  assertTrue(dynContains(hdls, "hdl3"));

  fwRPC_msgBroker_getHandlesCalledBy(_clientname, "p", hdls);
  assertEqual(1, dynlen(hdls));
  assertTrue(dynContains(hdls, "hdl4"));
}


void test_fwRPC_msgBroker_detects_new_mbx() {
  fwRPC_msgBroker_initialize();
  fwRPC_msgBroker_listen(true, "_mockNewMbxCB");

  delay(0, 500);

  //explicitly trigger new mbx signal
  string dp = _rbDp + ".initTrigger";
  dpSetWait(dp, true);
  dpSetWait(dp, false);
  
  //this should be one but the system may have multiple mbx already
  assertTrue(dynlen(_msgs) >= 1, "should detect msgbox");
  string proj = fwRPC_common_cleanProjectName(getSystemName());

  bool found = false;
  for (int i=1; i<=dynlen(_msgs); ++i) {
    if (_msgs[i][1] == proj && _msgs[i][2] == "loc_clnt_x") {
      found = true;
      break;
    }
  }

  assertTrue(found, "detected msg box should be the one expected");
}
